/* matmult-thread.c - Thread-Variante von matmult.c.   */
/* Copyright (C) 2000 by Ralph Schlosser, alle Rechte vorbehalten. */
/* Datum der Erstellung: 27062000     */
/* e-mail: ralph.schlosser@informatik.tu-chemnitz.de   */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include "pthread.h"

#define MAX     1000 /* 3000 */
#define DEBUG  0
#define MAX_THREADS 4

void createmat(void);
void *multiplythread(void *arg);

#if DEBUG == 1
void printmat(int *mt[MAX]);
#endif

/* GLOBALE VARIABLEN */

/* Zeiger fuer Matrizenstruktur */
/* Diese werden als "volatile" definiert, d.h. sie liegen wirklich */
/* im Speicher. Dies sollte generell fuer alle Daten verwendet */
/* werden, auf die verschiedene Threads schreiben und lesen. */
volatile int **a, **b, **c;
volatile int *temp1, *temp2, *temp3;

/* Groesse der Matrizen */
int size;

/* Anzahl der Threads */
int threads;

int main(int argc, char **argv)
{
	double teff1, teff2;
	long t1, t2, tm1, tm2;
	int t, i;
	int stat = 0;

	/* Array fuer Thread-Handles */
	pthread_t thread_handle[MAX_THREADS];

	/* Array fuer Thread-IDs */
	int thread_id[MAX_THREADS];

	void *retval;
	struct timeval tv;

	if (argc < 3) {
		printf("Fehlende Parameter!\n");
		printf("Benutzung: %s <Threads> <Groesse>\n", argv[0]);
		printf("z.B.: %s 4 500\n", argv[0]);
		exit(1);
	} else {
		threads = atoi(argv[1]);
		if (threads > MAX_THREADS) {
			printf("zu viele Threads, verwende Maximalanzahl (%d)\n", MAX_THREADS);
			threads = MAX_THREADS;
		}
		size = atoi(argv[2]);
		if (size > MAX) {
			printf("zu gross, verwende Maximalwert (%d)\n", MAX);
			size = MAX;
		} else printf("verwendete Groesse: %d\n", size);
	}

	printf("Erstelle zwei (%d x %d)-Matrizen\n", size, size);
	fflush(stdout);
	createmat();
	printf("Fertig. Bitte Taste drcken...\n");
	t = getchar();

	/* Thread-IDs initialisieren */
	for (i = 0; i < threads; i++) {
		thread_id[i] = i;
	}

	/* Start-Zeit nehmen */
	gettimeofday(&tv, NULL);
	t1 = tv.tv_sec;
	tm1 = tv.tv_usec;

	/* Threads starten (so viele wie gewuenscht).*/
	for (i = 0; i < threads; i++) {
		stat = stat || pthread_create(&thread_handle[i], NULL, multiplythread, &thread_id[i]);
	}
	if (stat) {
		fprintf(stderr, "%s: Probleme mit Threaderzeugung!\n", argv[0]);
		exit(1);
	}

	/* Warten bis alle Threads beendet sind.*/
	for (i = 0; i < threads; i++) {
		stat = stat || pthread_join(thread_handle[i], &retval);
	}
	if (stat) {
		fprintf(stderr, "%s: Konnte Threads nicht synchronisieren.\n", argv[0]);
		exit(1);
	}

	/*
	if(pthread_create(&thread0, NULL, multiplythread, &id_0) ||
	   pthread_create(&thread1, NULL, multiplythread, &id_1) ||
	   pthread_create(&thread2, NULL, multiplythread, &id_2) ||
	   pthread_create(&thread3, NULL, multiplythread, &id_3))
	{
	   fprintf(stderr, "%s: Konnte Thread nicht erzeugen.\n", argv[0]);
	   exit(1);
	}

	if(pthread_join(thread0, &retval) ||
	   pthread_join(thread1, &retval) ||
	   pthread_join(thread2, &retval) ||
	   pthread_join(thread3, &retval)) 
	{
	   fprintf(stderr, "%s: Konnte Threads nicht synchronisieren.\n", argv[0]);
	   exit(1);
	}
	*/

	/* Stop-Zeit nehmen */
	gettimeofday(&tv, NULL);
	t2 = tv.tv_sec;
	tm2 = tv.tv_usec;

	teff1 = t1 * 1e6 + tm1;
	teff2 = t2 * 1e6 + tm2;

	printf("Zeit fr die Berechnung in Sekunden: %f\n", (teff2 - teff1) / 1e6);

#if DEBUG == 1
	printf("Matrix A:\n");
	printmat(a);
	printf("Matrix B:\n");
	printmat(b);
	printf("Ergebnismatrix C:\n");
	printmat(c);
#endif
	return(0);
}

void createmat(void)
{
	int i, j;

	printf("Allokiere Speicher...\n");

	a = (volatile int **)malloc(sizeof(int *) * size);
	b = (volatile int **)malloc(sizeof(int *) * size);
	c = (volatile int **)malloc(sizeof(int *) * size);
	temp1 = (volatile int *)malloc(sizeof(int) * size * size);
	temp2 = (volatile int *)malloc(sizeof(int) * size * size);
	temp3 = (volatile int *)malloc(sizeof(int) * size * size);

	for (i = 0; i < size; i++)
		a[i] = &temp1[i * size];
	for (i = 0; i < size; i++)
		b[i] = &temp2[i * size];
	for (i = 0; i < size; i++)
		c[i] = &temp3[i * size];

	printf("Flle Speicher...\n");

	srand(time(NULL));
	for (i = 0; i < size; i++) {
		for (j = 0; j < size; j++) {
			a[i][j] = 1 + (int)((float)10 * rand() / (RAND_MAX + 1.0));
			b[i][j] = 1 + (int)((float)10 * rand() / (RAND_MAX + 1.0));
			c[i][j] = 0;
		}
	}
}

#if DEBUG == 1
void printmat(int *mat[size])
{
	int i, j;

	printf("\n");
	for (i = 0; i < size; i++) {
		for (j = 0; j < size; j++) {
			printf("%d ", mat[i][j]);
		}
		printf("\n");
	}
}
#endif

void *multiplythread(void *arg)
{
	register int start = *((int *)arg);
	int i, j, l, sum;

#if DEBUG == 1
	printf("Starting at: %d\n", start);
#endif
	for (i = start; i < size; i += threads) {
		for (j = 0; j < size; j++) {
			sum = 0;

			for (l = 0; l < size; l++) {
				sum += a[i][l] * b[l][j];
			}

			c[i][j] = sum;
		}
	}

	return(NULL);
}
