/* matmult.c - Multipliziert zwei (n x n)-Matrizen miteinander.  */
/* Copyright (C) 2000 by Ralph Schlosser, alle Rechte vorbehalten. */
/* Serielle Variante.       */
/* Datum der Erstellung: 23062000     */
/* Datum der letzten �derung: 29062000     */
/* e-mail: ralph.schlosser@informatik.tu-chemnitz.de   */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>

#define MAX 1000
#define DEBUG 0

void createmat(void);
void multiply(void);

#if DEBUG == 1
//void printmat(int *mt[size]);
void printmat(int *mt[MAX]);
#endif

volatile int **a, **b, **c;
volatile int *temp1, *temp2, *temp3;

volatile int size;

int main(int argc, char **argv)
{
	double teff1, teff2;
	long t1, t2, tm1, tm2;
	int t;
	struct timeval tv;

	if (argc < 2) {
		printf("Keine Groesse angegeben, verwende Default (%d)\n", MAX);
		size = MAX;
	} else {
		size = atoi(argv[1]);
		if (size > MAX) {
			printf("zu gross, verwende Maximalwert (%d)\n", MAX);
			size = MAX;
		} else printf("verwendete Groesse: %d\n", size);
	}

	printf("Erstelle Matrizen...\n");
	createmat();
	printf("Fertig. Bitte Taste drcken...\n");
	t = getchar();

	gettimeofday(&tv, NULL);
	t1 = tv.tv_sec;
	tm1 = tv.tv_usec;

	multiply();

	gettimeofday(&tv, NULL);
	t2 = tv.tv_sec;
	tm2 = tv.tv_usec;

	teff1 = t1 * 1e6 + tm1;
	teff2 = t2 * 1e6 + tm2;

#if DEBUG == 1
	printf("Matrix A:\n");
	printmat(a);
	printf("Matrix B:\n");
	printmat(b);
	printf("Ergebnismatrix C:\n");
	printmat(c);
#endif

	printf("Zeit fr die Berechnung in Sekunden: %f\n", (teff2 - teff1) / 1e6);

	return(0);
}

void createmat(void)
{
	int i, j;

	printf("Allokiere Speicher...\n");

	a = (volatile int **)malloc(sizeof(int *) * size);
	b = (volatile int **)malloc(sizeof(int *) * size);
	c = (volatile int **)malloc(sizeof(int *) * size);
	temp1 = (volatile int *)malloc(sizeof(int) * size * size);
	temp2 = (volatile int *)malloc(sizeof(int) * size * size);
	temp3 = (volatile int *)malloc(sizeof(int) * size * size);

	for (i = 0; i < size; i++)
		a[i] = &temp1[i * size];
	for (i = 0; i < size; i++)
		b[i] = &temp2[i * size];
	for (i = 0; i < size; i++)
		c[i] = &temp3[i * size];

	printf("Flle Speicher\n");

	srand(time(NULL));
	for (i = 0; i < size; i++) {
		for (j = 0; j < size; j++) {
			a[i][j] = 1 + (int)((float)10 * rand() / (RAND_MAX + 1.0));
			b[i][j] = 1 + (int)((float)10 * rand() / (RAND_MAX + 1.0));
			c[i][j] = 0;
		}
	}
}

#if DEBUG == 1
//void printmat(int *mat[size])
void printmat(int *mat[MAX])
{
	int i, j;

	printf("\n");
	for (i = 0; i < size; i++) {
		for (j = 0; j < size; j++) {
			printf("%d ", mat[i][j]);
		}
		printf("\n");
	}
}
#endif

void multiply(void)
{
	int i, j, l, sum;

	for (i = 0; i < size; i++) {
		for (j = 0; j < size; j++) {
			sum = 0;

			for (l = 0; l < size; l++) {
				sum += a[i][l] * b[l][j];
			}

			c[i][j] = sum;
		}
	}
}
