#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include "pthread.h"

#define MAX     10000000
#define DEBUG  0
#define MAX_THREADS 4

#define FLOAT float

volatile int size;

void createvec(void);
void *normalizethread(void *arg);
void check_length(FLOAT *vec);

#if DEBUG == 1
void printvec(FLOAT *vec);
#endif

volatile FLOAT *input_vec, *result_vec;

/* Array fuer die partielle Summe der einzelnen Threads. */
volatile FLOAT partial_sum[MAX_THREADS];

/* Array fuer den Barrier.*/
volatile int   barrier_status[MAX_THREADS];

int threads;

int main(int argc, char **argv)
{
	double teff1, teff2;
	long t1, t2, tm1, tm2;
	int t, i;
	pthread_t thread_handle[MAX_THREADS];
	int thread_id[MAX_THREADS];
	int stat = 0;
	void *retval;
	struct timeval tv;

	if (argc < 3) {
		printf("Fehlende Parameter!\n");
		printf("Benutzung: %s <threads> <Groesse>\n", argv[0]);
		printf("z.B.: %s 4 500000\n", argv[0]);
		exit(1);
	} else {
		threads = atoi(argv[1]);
		if (threads > MAX_THREADS) {
			printf("zu viele Threads, verwende Maximalanzahl (%d)\n", MAX_THREADS);
			threads = MAX_THREADS;
		}
		size = atoi(argv[2]);
		if (size > MAX) {
			printf("zu gross, verwende Maximalwert (%d)\n", MAX);
			size = MAX;
		} else printf("verwendete Groesse: %d\n", size);
	}

	printf("Erstelle Eingangs-Vektor (%d Komponenten)\n", size);
	fflush(stdout);
	createvec();
	printf("Fertig. Bitte Taste druecken...\n");
	t = getchar();

	/* Barrier Status ruecksetzen und Thread-IDs initialisieren */
	for (i = 0; i < threads; i++) {
		barrier_status[i] = 0;
		thread_id[i] = i;
	}

	gettimeofday(&tv, NULL);
	t1 = tv.tv_sec;
	tm1 = tv.tv_usec;


	/* Starten der Threads */
	for (i = 0; i < threads; i++) {
		stat = stat || pthread_create(&thread_handle[i], NULL, normalizethread, &thread_id[i]);
	}
	if (stat) {
		fprintf(stderr, "%s: Probleme mit Threaderzeugung!\n", argv[0]);
		exit(1);
	}


	/* Synchronisieren (Joinen) der Threads */
	for (i = 0; i < threads; i++) {
		stat = stat || pthread_join(thread_handle[i], &retval);
	}
	if (stat) {
		fprintf(stderr, "%s: Konnte Threads nicht synchronisieren.\n", argv[0]);
		exit(1);
	}


	gettimeofday(&tv, NULL);
	t2 = tv.tv_sec;
	tm2 = tv.tv_usec;

	teff1 = t1 * 1e6 + tm1;
	teff2 = t2 * 1e6 + tm2;

	printf("Zeit fuer die Berechnung in Sekunden: %f\n", (teff2 - teff1) / 1e6);

	check_length((FLOAT *) result_vec);

#if DEBUG == 1
	printf("Eingangsvector:\n");
	printvec((FLOAT *) input_vec);
	printf("Ergebnisvector:\n");
	printvec((FLOAT *) result_vec);
#endif
	return(0);
}

void createvec(void)
{
	int i;
	printf("Allokiere Speicher...\n");

	input_vec  = (volatile FLOAT *) malloc(sizeof(FLOAT) * size);
	result_vec = (volatile FLOAT *) malloc(sizeof(FLOAT) * size);

	printf("Fuelle Speicher...\n");

	for (i = 0; i < size; i++) {
		input_vec[i]  = 10 * rand() / (RAND_MAX + 1.0);
		result_vec[i] = 0;
	}
}


void *normalizethread(void *arg)
{
	register int my_ID = *((int *)arg);
	int i;
	int barrier_test;
	FLOAT vec_length = 0;
	FLOAT sum = 0;

	/* Berechnen der Summe der Quadrate "meiner" Komponenten */
	for (i = my_ID; i < size; i += threads) {
		sum += input_vec[i] * input_vec[i];
	}
	partial_sum[my_ID] = sum;

	/* Barrier */
	barrier_status[my_ID] = 1;

	do {
		barrier_test = 0;
		for (i = 0; i < threads; i++) {
			barrier_test += barrier_status[i];
		}
	} while (barrier_test != threads);

	/* Addition der Teilsummen und entgueltige Laengenberechnung */
	for (i = 0; i < threads; i++) {
		vec_length += partial_sum[i];
	}
	vec_length = 1.0 / sqrt(vec_length);

	/* Skalierung "meiner" Komponenten */
	for (i = my_ID; i < size; i += threads) {
		result_vec[i] = input_vec[i] * vec_length;
	}

	return 0;
}

void check_length(FLOAT *vec)
{
	int i;
	FLOAT sum = 0;
	FLOAT length;

	for (i = 0; i < size; i++) {
		sum += vec[i] * vec[i];
	}
	length = sqrt(sum);
	printf("Vectorlaenge: %f\n", length);
}

#if DEBUG == 1
void printvec(FLOAT *vec)
{
	int i;

	for (i = 0; i < size; i++) {
		printf("%f ", vec[i]);
	}
	printf("\n");
}
#endif
