#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include "pthread.h"

#define MAX     10000000 /* 3000 */
#define DEBUG  0

int size;

void createvec(void);
void normalize();
void check_length(float *vec);

#if DEBUG == 1
void printvec(float *vec);
#endif

volatile float *input_vec, *result_vec;

int main(int argc, char **argv)
{
	double teff1, teff2;
	long t1, t2, tm1, tm2;
	int t;
	struct timeval tv;

	if (argc < 2) {
		printf("Fehlende Parameter!\n");
		printf("Benutzung: %s <Groesse>\n", argv[0]);
		printf("z.B.: %s 500000\n", argv[0]);
		exit(1);
	} else {
		size = atoi(argv[1]);
		if (size > MAX) {
			printf("zu gross, verwende Maximalwert (%d)\n", MAX);
			size = MAX;
		} else printf("verwendete Groesse: %d\n", size);
	}

	printf("Erstelle Eingangs-Vektor (%d Komponenten)\n", size);
	fflush(stdout);
	createvec();
	printf("Fertig. Bitte Taste drcken...\n");
	t = getchar();

	gettimeofday(&tv, NULL);
	t1 = tv.tv_sec;
	tm1 = tv.tv_usec;

	normalize();

	gettimeofday(&tv, NULL);
	t2 = tv.tv_sec;
	tm2 = tv.tv_usec;

	teff1 = t1 * 1e6 + tm1;
	teff2 = t2 * 1e6 + tm2;

	check_length((float *) result_vec);

	printf("Zeit fr die Berechnung in Sekunden: %f\n", (teff2 - teff1) / 1e6);

#if DEBUG == 1
	printf("Eingangsvector:\n");
	printvec((float *) input_vec);
	printf("Ergebnisvector:\n");
	printvec((float *) result_vec);
#endif
	return(0);
}

void createvec(void)
{
	int i;
	printf("Allokiere Speicher...\n");

	input_vec  = (volatile float *) malloc(sizeof(float) * size);
	result_vec = (volatile float *) malloc(sizeof(float) * size);

	printf("Flle Speicher...\n");

	for (i = 0; i < size; i++) {
		input_vec[i]  = (float) rand();
		result_vec[i] = 0;
	}
}

void normalize()
{
	int i;
	float vec_length = 0;

	/* zunaechst Laenge des Vectors ermitteln */
	for (i = 0; i < size; i++) {
		vec_length = vec_length + input_vec[i] * input_vec[i] ;
	}
	vec_length = sqrt(vec_length);

	/* jetzt skalieren */
	for (i = 0; i < size; i++) {
		result_vec[i] = input_vec[i] / vec_length;
	}
}

void check_length(float *vec)
{
	int i;
	float sum = 0;
	float length;

	for (i = 0; i < size; i++) {
		sum += vec[i] * vec[i];
	}
	length = sqrt(sum);
	printf("Vectorlaenge: %f\n", length);
}

#if DEBUG == 1
void printvec(float *vec)
{
	int i;

	for (i = 0; i < size; i++) {
		printf("%f ", vec[i]);
	}
	printf("\n");
}
#endif
